{
    Respond, the fax number form daemon
    Copyright (C) 2000 Horst F <horstf@gmx.de>

    This program is free software; you can redistribute it and/or modify
    it under the terms of the GNU General Public License as published by
    the Free Software Foundation; either version 2 of the License, or
    (at your option) any later version.

    This program is distributed in the hope that it will be useful,
    but WITHOUT ANY WARRANTY; without even the implied warranty of
    MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
    GNU General Public License for more details.

    You should have received a copy of the GNU General Public License
    along with this program; if not, write to the Free Software
    Foundation, Inc., 59 Temple Place, Suite 330, Boston, MA  02111-1307  USA
}
unit CommonDataAccess;

interface

uses
  Classes, SysUtils, IniFile, IniFiles, Registry, Observer,
  DatabaseDescription, TableDescription, ObservedStringList;

type
  {
    This class provides a uniform access to private ini files and the win.ini
    file. Other sources are possible (like database access).
    The class provides access for key-value pairs in a table and real database
    like access to fields of a string keyed record in a table. For a ini file
    database the term "table" means a section of the ini file.
    It is possible to define an alias for a table. In this case the table is
    only accessibly with the alias name, but the methods work with the real
    name. This allows accessing table with the same name in different databases.
    All access methods are static, i.e. class methods.
    No errors will be reported and no exception will be thrown. Illegal
    operations will not be executed. I think this is the way ini files (where
    the roots of TCommonData are) work.
  }
  TCommonData = class(TObject)
  private
    class function GetTableDatabase(const table: String): TDatabaseDescription;
    class function GetTableDescription(const table: String): TTableDescription;
  public
    class procedure BindTableToIniFile(const table, name: String);
    class procedure SetTableAlias(const table, alias: String);
    class procedure ReleaseTable(const table: String);
    class procedure DeleteKey(const table, key: String);
    class procedure EraseTable(const table: string);
    class procedure ReadTableKeys(const table: String; keys: TStrings);
    class procedure ReadTables(tablelist: TStrings);
    class procedure ReadTableValues(const table: String; values: TStrings);
    class function ReadInteger(const table, key: String; const Def: Integer)
        : Integer;
    class function ReadString(const table, key, def: String): String;
    class procedure ReadFields(const table, key: String; fields: TStrings);
    class procedure WriteFields(const table, key: String; fields: TStrings);
    class procedure WriteInteger(const table, key: String;
        const value: Integer);
    class procedure WriteString(const table, key, value: String);
    class procedure DefineFields(const table: String; const fields: TStrings);

    class function ReadIntegerField(const table, key, field: String;
        const Def: Integer): Integer;
    class function ReadStringField(const table, key, field, def: String)
        : String;
    class procedure WriteIntegerField(const table, key, field: String;
        const value: Integer);
    class procedure WriteStringField(const table, key, field, value: String);
    class procedure Flush(const table: String);
  end;

implementation

  var
    {
      Holds all opened databases.
    }
    oDatabaseDescriptions : TObservedStringList;

    {
      Holds all table-database assignments.
    }
    oTableDescriptions : TObservedStringList;

    {
      Saved unit exit procedure pointer.
    }
    SavedExitProc : Pointer;

  {
    Creates a connection to a file based database (currently only ini files).
    Ini files contain sections. A connection will be established in a kind,
    that a later access to the table will be an access to a section with the
    table name.
    If more then one table connection were made to one file, the file is opened
    only once.

    @param table tablename, which is to connect to the file
    @param name filename
  }
  class procedure TCommonData.BindTableToIniFile(const table, name: String);
  var
    atype : TCommonDatabaseType;
    dbdesc : TDatabaseDescription;
    tabledesc : TTableDescription;
    i, j : Integer;
    ini : String;
    db : TObject;
  begin
    if Length(Trim(table)) = 0 then exit;
    // Ignore more then one definition for a table
    if oTableDescriptions.Find(AnsiLowerCase(table), i) then exit;
    // Empty name means win.ini
    if (Length(Trim(name)) = 0) or (AnsiLowerCase(name) = 'win.ini') then begin
      atype := isWinIni;
      ini := 'win.ini';
    end else begin
      atype := isIniFile;
      ini := name;
    end;
    db := nil;
    try
      // Search for an existing database...
      j := -1;
      if oDatabaseDescriptions.Find(AnsiLowerCase(ini), i) then begin
        dbdesc := TDatabaseDescription(oDatabaseDescriptions.Objects[i]);
      end else begin
        // ... and build a new one if none with the name found
        if atype = isIniFile then begin
          db := TPrivateIniFile.Create(name)
        end else begin
          db := TIniFile.Create('WIN.INI');
        end;
        dbdesc := TDatabaseDescription.Create(ini, db, atype);
        try
          // Add the description to the database collection
          // Save index for possible deletion on error with table description
          j := oDatabaseDescriptions.AddObject(AnsiLowerCase(dbdesc.Name),
              dbdesc);
        except
          on e: EOutOfMemory do begin
            dbdesc.Free;
            raise;
          end;
        end;
      end;
      // Make new table description
      try
        tabledesc := TTableDescription.Create(table, dbdesc);
        try
          // Add it to the table database collection
          oTableDescriptions.AddObject(AnsiLowerCase(tabledesc.Name), tabledesc);
        except
          on EOutOfMemory do begin
            tabledesc.Free;
            raise;
          end;
        end;
        dbdesc.Link;
      except
        on EOutOfMemory do begin
          // Could not add table description to collection; if database was
          // opened right now, close it
          if j >= 0 then begin
            oDatabaseDescriptions.Delete(j);
            dbdesc.Free;
          end;
          raise;
        end;
      end;
    except
      // If an ini file object was created, destroy it on error
      on EOutOfMemory do if db <> nil then db.Free;
    end;
  end;

  {
    Changes the logical name of a table.
    The physical name (e.g. the section name) will not be changed. This allows
    the access to tables with the same name in different databases.

    @param table old name
    @param alias new name
  }
  class procedure TCommonData.SetTableAlias(const table, alias: String);
  var
    i : Integer;
    tab : TTableDescription;
  begin
    if Length(Trim(alias)) = 0 then exit;
    if oTableDescriptions.Find(AnsiLowerCase(alias), i) then exit;
    if oTableDescriptions.Find(AnsiLowerCase(table), i) then begin
      tab := TTableDescription(oTableDescriptions.Objects[i]);
      oTableDescriptions.Delete(i);
      try
        oTableDescriptions.AddObject(AnsiLowerCase(alias), tab);
      except
        on EOutOfMemory do;
      end;
    end;
  end;

  {
    Returns the database description for the table.

    @param table tablename
  }
  class function TCommonData.GetTableDatabase(const table: String)
      : TDatabaseDescription;
  var
    i : Integer;
  begin
    Result := nil;
    if oTableDescriptions.Find(AnsiLowerCase(table), i) then begin
      Result := TTableDescription(oTableDescriptions.Objects[i]).DatabaseDescription;
    end;
  end;

  {
    Returns the table description.

    @param table tablename
    @return the table description
  }
  class function TCommonData.GetTableDescription(const table: String)
      : TTableDescription;
  var
    i : Integer;
  begin
    Result := nil;
    if oTableDescriptions.Find(AnsiLowerCase(table), i) then begin
      Result := TTableDescription(oTableDescriptions.Objects[i]);
    end;
  end;

  {
    Releases a table.
    The table is no longer accessible. The database is freed only if no other
    table is connected to it. An ini file is closed when no more tables
    (sections) are used.

    @param table tablename
  }

  class procedure TCommonData.ReleaseTable(const table: String);
  var
    i : Integer;
    dbdesc : TDatabaseDescription;
    tabdesc : TTableDescription;
    db : TObject;
  begin
    if oTableDescriptions.Find(AnsiLowerCase(table), i) then begin
      // Remove table fom table description list
      tabdesc := TTableDescription(oTableDescriptions.Objects[i]);
      oTableDescriptions.Delete(i);
      dbdesc := tabdesc.DatabaseDescription;
      tabdesc.Free;
      dbdesc.Unlink;
      // Close database if no more links
      if not dbdesc.IsLinked then begin
        if oDatabaseDescriptions.Find(AnsiLowerCase(dbdesc.Name), i) then begin
          oDatabaseDescriptions.Delete(i);
        end;
        db := dbdesc.Database;
        dbdesc.Free;
        db.Free;
      end;
    end;
  end;

  {
    Deletes a key from a table.

    @param table tablename
    @param key key to delete
  }
  class procedure TCommonData.DeleteKey(const table, key: String);
  var
    db : TDatabaseDescription;
    tab : TTableDescription;
  begin
    tab := GetTableDescription(table);
    if tab = nil then exit;
    db := tab.DatabaseDescription;
    case db.DatabaseType of
      isIniFile: begin
        TPrivateIniFile(db.Database).DeleteKey(tab.Name, key);
      end;
      isWinIni: begin
        TIniFile(db.Database).DeleteKey(tab.Name, key);
      end;
    end;
  end;

  {
    Deletes a table from the database.
    The table is released and if there are no more used tables in the database,
    it is closed.

    @param table tablename
  }
  class procedure TCommonData.EraseTable(const table: string);
  var
    dbdesc : TDatabaseDescription;
    db : TObject;
    tab : TTableDescription;
    i : Integer;
  begin
    if oTableDescriptions.Find(AnsiLowerCase(table), i) then begin
      // Remove table fom table description list
      tab := TTableDescription(oTableDescriptions.Objects[i]);
      oTableDescriptions.Delete(i);
      dbdesc := tab.DatabaseDescription;
      dbdesc.Unlink;
      // Erase table from database
      db := dbdesc.Database;
      case dbdesc.DatabaseType of
        isIniFile: begin
          TPrivateIniFile(db).EraseSection(tab.Name);
        end;
        isWinIni: begin
          TIniFile(db).EraseSection(tab.Name);
        end;
      end;
      tab.Free;
      // Close database if no more links
      if not dbdesc.IsLinked then begin
        if oDatabaseDescriptions.Find(AnsiLowerCase(dbdesc.Name), i) then begin
          oDatabaseDescriptions.Delete(i);
        end;
        dbdesc.Free;
        db.Free;
      end;
    end;
  end;

  {
    Reads all keys from a table of the database.

    @param table tablename
    @param keys collection to write the keys to
  }
  class procedure TCommonData.ReadTableKeys(const table: String; keys: TStrings);
  var
    db : TDatabaseDescription;
    tab : TTableDescription;
  begin
    tab := GetTableDescription(table);
    if tab = nil then exit;
    db := tab.DatabaseDescription;
    case db.DatabaseType of
      isIniFile: begin
        TPrivateIniFile(db.Database).ReadSection(tab.Name, keys);
      end;
      isWinIni: begin
        TIniFile(db.Database).ReadSection(tab.Name, keys);
      end;
    end;
  end;

  {
    Reads all table names from all opened databases.
    This will not work correctly with table aliases!!!

    @param tablelist collection to write the names to
  }
  class procedure TCommonData.ReadTables(tablelist: TStrings);
  var
    db : TDatabaseDescription;
    i : Integer;
  begin
    for i := 0 to oDatabaseDescriptions.Count - 1 do begin
      db := TDatabaseDescription(oDatabaseDescriptions.Objects[i]);
      case db.DatabaseType of
        isIniFile: begin
          TPrivateIniFile(db.Database).ReadSections(tablelist);
        end;
        isWinIni: begin
          TIniFile(db.Database).ReadSections(tablelist);
        end;
      end;
    end;
  end;

  {
    Reads all values from a table of the database.

    @param table tablename
    @param values collection to write the values to
  }
  class procedure TCommonData.ReadTableValues(const table: String;
      values: TStrings);
  var
    db : TDatabaseDescription;
    tab : TTableDescription;
  begin
    tab := GetTableDescription(table);
    if tab = nil then exit;
    db := tab.DatabaseDescription;
    case db.DatabaseType of
      isIniFile: begin
        TPrivateIniFile(db.Database).ReadSectionValues(tab.Name, values);
      end;
      isWinIni: begin
        TIniFile(db.Database).ReadSectionValues(tab.Name, values);
      end;
    end;
  end;

  {
    Reads an integer value from a table of the database.

    @param table tablename
    @param key key of the value to read
    @param def default value

    @return the requested value
  }
  class function TCommonData.ReadInteger(const table, key: String;
      const def: Integer): Integer;
  var
    db : TDatabaseDescription;
    tab : TTableDescription;
  begin
    Result := def;
    tab := GetTableDescription(table);
    if tab = nil then exit;
    db := tab.DatabaseDescription;
    case db.DatabaseType of
      isIniFile: begin
        Result :=
            TPrivateIniFile(db.Database).ReadInteger(tab.Name, key, def);
      end;
      isWinIni: begin
        Result := TIniFile(db.Database).ReadInteger(tab.Name, key, def);
      end;
    end;
  end;

  {
    Reads a string from a table of the database.

    @param table tablename
    @param key key of the value to read
    @param def default value

    @return the requested value
  }
  class function TCommonData.ReadString(const table, key, def: String): String;
  var
    db : TDatabaseDescription;
    tab : TTableDescription;
  begin
    Result := def;
    tab := GetTableDescription(table);
    if tab = nil then exit;
    db := tab.DatabaseDescription;
    case db.DatabaseType of
      isIniFile: begin
        Result :=
            TPrivateIniFile(db.Database).ReadString(tab.Name, key, def);
      end;
      isWinIni: begin
        Result := TIniFile(db.Database).ReadString(tab.Name, key, def);
      end;
    end;
  end;

  {
    Writes an integer value to a table.

    @param table tablename
    @param key key of the value to write
    @param value the value to write
  }
  class procedure TCommonData.WriteInteger(const table, key: String;
      const value: Integer);
  var
    db : TDatabaseDescription;
    tab : TTableDescription;
  begin
    tab := GetTableDescription(table);
    if tab = nil then exit;
    db := tab.DatabaseDescription;
    case db.DatabaseType of
      isIniFile: begin
        TPrivateIniFile(db.Database).WriteInteger(tab.Name, key, value);
      end;
      isWinIni: begin
        TIniFile(db.Database).WriteInteger(tab.Name, key, value);
      end;
    end;
  end;

  {
    Writes a string to a table of the database.

    @param table tablename
    @param key key of the value to write
    @param value the value to write
  }
  class procedure TCommonData.WriteString(const table, key, value: String);
  var
    db : TDatabaseDescription;
    tab : TTableDescription;
  begin
    tab := GetTableDescription(table);
    if tab = nil then exit;
    db := tab.DatabaseDescription;
    case db.DatabaseType of
      isIniFile: begin
        TPrivateIniFile(db.Database).WriteString(tab.Name, key, value);
      end;
      isWinIni: begin
        TIniFile(db.Database).WriteString(tab.Name, key, value);
      end;
    end;
  end;

  {
    Defines a field list for a table or section of the database.
    This is need prior using field based access methods.

    @param table tablename
    @param fields collection of the field names
    @param separator the separator string for string based databases like
                     ini files
  }
  class procedure TCommonData.DefineFields(const table: String;
      const fields: TStrings);
  var
    tab : TTableDescription;
  begin
    tab := GetTableDescription(table);
    if tab <> nil then tab.Fields := fields;
  end;

  {
    Reads an integer value from a table of the database field based.

    @param table tablename
    @param key key of the record of the field to read
    @param field field name of the value to read
    @param def default value

    @return the requested value
  }
  class function TCommonData.ReadIntegerField(const table, key, field: String;
      const def: Integer): Integer;
  var
    s: String;
  begin
    s := ReadStringField(table, key, field, IntToStr(def));
    try
      Result := StrToInt(s);
    except
      on EConvertError do Result := def;
    end;
  end;

  {
    Reads a string from a table of the database field based.

    @param table tablename
    @param key key of the record of the field to read
    @param field field name of the value to read
    @param def default value

    @return the requested value
  }
  class function TCommonData.ReadStringField(const table, key, field,
      def: String): String;
  var
    db : TDatabaseDescription;
    tab : TTableDescription;
    i, j : Integer;
    nl : Boolean;
    s: String;
  begin
    Result := def;
    tab := GetTableDescription(table);
    if tab = nil then exit;
    db := tab.DatabaseDescription;
    if db.DatabaseType in [isIniFile, isWinIni] then begin
      if tab.Fields = nil then exit;
      i := tab.Fields.IndexOf(field);
      // Exit if no such field
      if i < 0 then exit;
      // Is this field not the last field?
      nl := i < tab.Fields.Count - 1;
      s := ReadString(table, key, '');
      // Step over the first not needed fields
      while i > 0 do begin
        j := Pos(tab.Separator, s);
        if j > 0 then begin
          Delete(s, 1, j + Length(tab.Separator));
        end else begin
          // Exit, if not enough separators
          exit;
        end;
        Dec(i);
      end;
      // Remove fields behind the needed, but only
      if nl then begin
        j := Pos(tab.Separator, s);
        if j > 0 then begin
          Delete(s, j, Length(s));
        end;
      end;
      Result := s;
    end;
  end;

  {
    Reads all values of a given key from a table.

    @param table tablename
    @param key key of the record of the field to read
    @param values collection to write the values to
  }
  class procedure TCommonData.ReadFields(const table, key: String;
      fields: TStrings);
  var
    s, sep : String;
    i, count : Integer;
    tab : TTableDescription;
  begin
    tab := GetTableDescription(table);
    if tab = nil then exit;
    if tab.DatabaseDescription.DatabaseType in [isIniFile, isWinIni] then begin
      if tab.Fields = nil then exit;
      s := ReadString(table, key, '');
      if Length(s) > 0 then begin
        count := tab.Fields.Count;
        fields.BeginUpdate;
        try
          while count > 0 do begin
            Dec(count);
            i := Pos(tab.Separator, s);
            if (i > 0) and (count > 0) then begin
              fields.Add(Copy(s, 1, i - 1));
              Delete(s, 1, i + Length(tab.Separator) - 1);
            end else begin
              if i = 0 then begin
                sep := Trim(tab.Separator);
                i := Pos(sep, s);
                if i = Length(s) - Length(sep) + 1 then begin
                  Delete(s, i, Length(sep));
                end;
              end;
              fields.Add(s);
              s := '';
            end;
          end;
        except
          on EOutOfMemory do;
        end;
        fields.EndUpdate;
      end;
    end;
  end;

  {
    Writes all values of a given key to a table.

    @param table tablename
    @param key key of the record of the field to read
    @param values collection to write the values to
  }
  class procedure TCommonData.WriteFields(const table, key: String;
      fields: TStrings);
  var
    s : String;
    i : Integer;
    tab : TTableDescription;
  begin
    tab := GetTableDescription(table);
    if tab = nil then exit;
    if tab.DatabaseDescription.DatabaseType in [isIniFile, isWinIni] then begin
      if tab.Fields = nil then exit;
      s := '';
      for i := 0 to fields.Count - 1 do begin
        if i > 0 then s := s + tab.Separator;
        s := s + fields.Strings[i];
      end;
      WriteString(table, key, s);
    end;
  end;

  {
    Writes an integer value to a table of the database field based.

    @param table tablename
    @param key key of the record of the field to write
    @param field field name of the value to write
    @param value value to write
  }
  class procedure TCommonData.WriteIntegerField(const table, key, field: String;
      const value: Integer);
  begin
    WriteStringField(table, key, field, IntToStr(value));
  end;

  {
    Writes a string to a table of the database field based.

    @param table tablename
    @param key key of the record of the field to write
    @param field field name of the value to write
    @param value value to write
  }
  class procedure TCommonData.WriteStringField(
      const table, key, field, value: String);
  var
    db : TDatabaseDescription;
    tab : TTableDescription;
    i, j : Integer;
    nl : Boolean;
    s, s1 : String;
  begin
    tab := GetTableDescription(table);
    if tab = nil then exit;
    db := tab.DatabaseDescription;
    if db.DatabaseType in [isIniFile, isWinIni] then begin
      if tab.Fields = nil then exit;
      i := tab.Fields.IndexOf(field);
      // Exit if no such field
      if i < 0 then exit;
      // Is this field not the last field?
      nl := i < tab.Fields.Count - 1;
      s := ReadString(table, key, '');
      s1 := '';
      // Step over the first not needed fields
      while i > 0 do begin
        j := Pos(tab.Separator, s);
        if j > 0 then begin
          s1 := s1 + Copy(s, 1, j + Length(tab.Separator));
          Delete(s, 1, j + Length(tab.Separator));
        end else begin
          // Exit, if not enough separators
          exit;
        end;
        Dec(i);
      end;
      s1 := s1 + value;
      // Get fields behind the needed
      if nl then begin
        j := Pos(tab.Separator, s);
        if j > 0 then begin
          s1 := s1 + Copy(s, j, Length(s));
        end else begin
          // It's not the last field, so append a separator
          s1 := s1 + tab.Separator;
        end;
      end;
    end;
  end;

  {
    Flushes the database, if possible.

    @param table tablename
  }
  class procedure TCommonData.Flush(const table: String);
  var
    db : TDatabaseDescription;
  begin
    db := GetTableDatabase(table);
    if db = nil then exit;
    // Currently only TPrivatIniFile supports flushing
    if db.DatabaseType in [isIniFile] then begin
      TPrivateIniFile(db.Database).Flush
    end;
  end;

  {
    The unit exit procedure.
  }
  procedure ExitProcedure;
  var
    i : Integer;
    db : TDatabaseDescription;
  begin
    ExitProc := SavedExitProc;
    // Free all table descriptions
    oTableDescriptions.BeginUpdate;
    for i := 0 to oTableDescriptions.Count - 1 do begin
      oTableDescriptions.Objects[i].Free;
    end;
    oTableDescriptions.EndUpdate;
    oTableDescriptions.Clear;
    oTableDescriptions.Free;
    // Free (close) all open databases
    oDatabaseDescriptions.BeginUpdate;
    for i := 0 to oDatabaseDescriptions.Count - 1 do begin
      db := TDatabaseDescription(oDatabaseDescriptions.Objects[i]);
      db.Database.Free;
      db.Free;
    end;
    oDatabaseDescriptions.EndUpdate;
    oDatabaseDescriptions.Clear;
    oDatabaseDescriptions.Free;
  end;

begin
  oDatabaseDescriptions := TObservedStringList.Create;
  with oDatabaseDescriptions do begin
    Sorted := true;
    Duplicates := dupError;  // Should not happen
  end;
  oTableDescriptions := TObservedStringList.Create;
  with oTableDescriptions do begin
    Sorted := True;
    Duplicates := dupIgnore;
  end;
  SavedExitProc := ExitProc;
  ExitProc := @ExitProcedure;
end.
