{
    Respond, the fax number form daemon
    Copyright (C) 2000 Horst F <horstf@gmx.de>

    This program is free software; you can redistribute it and/or modify
    it under the terms of the GNU General Public License as published by
    the Free Software Foundation; either version 2 of the License, or
    (at your option) any later version.

    This program is distributed in the hope that it will be useful,
    but WITHOUT ANY WARRANTY; without even the implied warranty of
    MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
    GNU General Public License for more details.

    You should have received a copy of the GNU General Public License
    along with this program; if not, write to the Free Software
    Foundation, Inc., 59 Temple Place, Suite 330, Boston, MA  02111-1307  USA
}
unit IniFile;

interface

uses
  classes, sysutils, forms, Observer, StringObject,
  ObservedStringList;

type
  {
    This class provides access to an windows ini file. Windows functions and
    TIniFile class are limited to 64kB, so to handle large phonebooks I wrote
    my own class.
  }
  TPrivateIniFile = class
  private
    // Contains section name as key and a string list for the section as object
    Sections : TObservedStringList;
    // File name
    oFileName: String;
    // True, if data should automatically be saved to file on destroy
    oPersistent: Boolean;
    // True, if a .bak file should be created in save
    oBackup: Boolean;
    // True, when data changed after create or last write
    oDirty: Boolean;
  public
    { TIniFile-like methods and properties }
    constructor Create(const name: String);
    destructor Destroy; override;
    procedure DeleteKey(const section, key: String);
    procedure EraseSection(const section: string);
    function ReadInteger(const section, key: String; const Def: Longint)
        : Longint;
    procedure ReadSection(const section: String; keys: TStrings);
    procedure ReadSections(sectlist: TStrings);
    procedure ReadSectionValues(const section: String; values: TStrings);
    function ReadString(const section, key, def: String): String;
    procedure WriteInteger(const section, key: String; const value: LongInt);
    procedure WriteString(const section, key, value: String);
    property FileName: String read oFileName;
    { Own Methods and properties }
    constructor CreateEmpty;
    procedure Clear;
    procedure Load(const name: String);
    procedure Save(const name: String);
    function putString(const section, key, value: String): Boolean;
    function putInt(const section, key: String; const value: Integer): Boolean;
    function CreateStringList: TObservedStringList;
    procedure Flush;
    property Persistent: Boolean read oPersistent write oPersistent;
    property Backup: Boolean read oBackup write oBackup;
    property Dirty: Boolean read oDirty;
  end;

implementation

  {
    Creates a sorted TObservedStringList object with error on duplicates.

    @return the new TObservedStringList object
  }
  function TPrivateIniFile.CreateStringList: TObservedStringList;
  var
    sl : TObservedStringList;
  begin
    try
      sl := TObservedStringList.Create;
      sl.Sorted := True;
      sl.Duplicates := dupError;
      Result := sl;
    except
      on EOutOfMemory do Result := nil;
    end;
  end;

  {
    Create the object and load data from a file.

    @param name the file name
  }
  constructor TPrivateIniFile.Create(const name: String);
  begin
    Sections := nil;
    oDirty := False;
    oPersistent := False;
    oBackup := True;
    SetLength(oFileName, 0);
    inherited Create;
    Sections := CreateStringList;
    Load(name);
    oPersistent := True;
    ObjectObserver.Observe(self);
  end;

  {
    Creates an empty object without loading data from a file.
  }
  constructor TPrivateIniFile.CreateEmpty;
  begin
    Sections := nil;
    oDirty := False;
    oPersistent := False;
    oBackup := True;
    SetLength(oFileName, 0);
    inherited Create;
    Sections := CreateStringList;
    ObjectObserver.Observe(self);
  end;

  {
    Destroys the object. If object is marked persistent and a
    filename is given, data will be saved to this file before destroying.
  }
  destructor TPrivateIniFile.Destroy;
  begin
    if oPersistent and (Length(oFileName) > 0) then Save(oFileName);
    if Sections <> nil then Clear;
    Sections.Free;
    ObjectObserver.Forget(self);
    inherited Destroy;
  end;

  {
    Removes all data from the object.
  }
  procedure TPrivateIniFile.Clear;
  var
    i, j : Integer;
    ts : TString;
    sl : TObservedStringList;
  begin
    if Sections.Count > 0 then oDirty := True;
    for i := 0 to Sections.Count-1 do begin
      sl := TObservedStringList(Sections.Objects[i]);
      sl.BeginUpdate;
      for j := 0 to sl.Count-1 do begin
        ts := TString(sl.Objects[j]);
        ts.Free;
        if (j mod 100) = 99 then Application.ProcessMessages;
      end;
      sl.EndUpdate;
      sl.Clear;
      sl.Free;
    end;
    Sections.Clear;
  end;

  {
    Loads the given ini file into the object.

    @param name the filename
  }
  procedure TPrivateIniFile.load(const name: String);
  var
    i, j : Integer;
    f : TextFile;
    sl : TObservedStringList;
    line : String;
    key, value : String;
    ts : TString;
  begin
    oFileName := Name;
    oDirty := True;
    {Open file}
    AssignFile(f, name);
    try
      try
        Reset(f);
        {On start we have no section open}
        sl := nil;
        j := 0;
        while not Eof(f) do begin
          Readln(f, line);
          line := Trim(line);
          {Ignore empty lines}
          if length(line) = 0 then continue;
          {Ignore comments}
          if line[1] = ';' then continue;
          {Does here a section start?}
          if line[1] = '[' then begin
            line := Copy(line, 2, Length(line)-1);
            {Malformed section name contains '['}
            if Pos('[', line) > 0 then continue;
            {Find end of section name}
            i := Pos(']', line);
            {There cannot be a name when ']' follows '[' immediatly}
            if i < 2 then continue;
            key := Trim(Copy(line, 1, i-1));
            {Ignore section names containing only spaces}
            if Length(key) = 0 then continue;
            {Try to find existing section name in section list}
            if Sections.Find(key, i) then begin
              {Get TObservedStringList for this section, if name was found}
              sl := TObservedStringList(Sections.Objects[i]);
            end else begin
              {Name was not found -> create a new TObservedStringList for this section}
              sl := CreateStringList;
              {If no memory stop loading}
              if sl = nil then break;
              try
                {Put the new section in the section list}
                Sections.addObject(key, sl);
              except
                on EOutOfMemory do begin
                  {If there is no memory, free the newly created list...}
                  sl.Free;
                  {...and stop loading}
                  break;
                end;
              end;
            end;
          end else begin
            {It's neither comment nor section name, must be key=value}
            {Ignore lines before the first section starts}
            if sl = nil then continue;
            i := Pos('=', line);
            {Keys must be at least one char long}
            if i < 2 then continue;
            key := Trim(Copy(line, 1, i-1));
            {This should be useless, only for paranoia ;)}
            if Length(key) = 0 then continue;
            value := Trim(Copy(line, i+1, Length(line)-i));
            // If the key exist, remove it and free the value
            if sl.Find(key, i) then begin
              ts := TString(sl.Objects[i]);
              sl.Delete(i);
              ts.free;
            end;
            ts := TString.Create(value);
            try
              {Add key-value-pair to section list}
              sl.addObject(key, ts);
            except
              on EOutOfMemory do begin
                {If there is no memory, free the dynamic string...}
                ts.Free;
                {...and stop loading}
                break;
              end;
              on EListError do ts.Free;
            end;
          end;
          Inc(j);
          if (j mod 100) = 99 then Application.ProcessMessages;
        end;
      finally
        CloseFile(f);
      end;
    except
      on EInOutError do;
    end;
  end;

  {
    Saves data to the given file.

    @param name the filename
  }
  procedure TPrivateIniFile.save(const name: String);
  var
    f : TextFile;
    s : String;
    i, j : Integer;
    sl : TObservedStringList;
    ts : TString;
  begin
    oFileName := name;
    // Backup if required
    if oBackup then begin
      {First remove old backup file}
      s := ChangeFileExt(name, '.BAK');
      DeleteFile(s);
      {Second rename old file to backup file}
      RenameFile(name, s);
    end;
    {Third write data to file}
    AssignFile(f, name);
    try
      try
        Rewrite(f);
        for i := 0 to Sections.Count-1 do begin
          // Get the section string list
          sl := TObservedStringList(Sections.Objects[i]);
          // Write section header, an empty line first
          Writeln(f);
          Writeln(f, '[' + Sections.Strings[i] + ']');
          for j := 0 to sl.Count-1 do begin
            // Get value
            ts := TString(sl.Objects[j]);
            // Write 'key=value'
            Writeln(f, sl.Strings[j] + '=' + ts.Value);
            // We won't block the system
            if (j mod 100) = 99 then Application.ProcessMessages;
          end;
        end;
      finally
        CloseFile(f);
      end;
      oDirty := False;
    except
      // On in/out-error restore the old file
      on EInOutError do begin
        DeleteFile(name);
        RenameFile(s, name);
      end;
    end;
  end;

  {
    Reads a string value.

    @param section section name
    @param key key name
    @param def default value, if key does not exist
  }
  function TPrivateIniFile.ReadString(const section, key, def: String): String;
  var
    sl : TObservedStringList;
    ts : TString;
    i : Integer;
  begin
    Result := '';
    if Sections.Find(section, i) then begin
      sl := TObservedStringList(Sections.Objects[i]);
      if sl.Find(key, i) then begin
        ts := TString(sl.Objects[i]);
        if ts <> nil then Result := ts.Value;
      end else Result := def;
    end else Result := def;
  end;

  {
    Reads an integer value.

    @param section section name
    @param key key name
    @param def default value, if key does not exist
  }
  function TPrivateIniFile.ReadInteger(const section, key: String;
      const def: LongInt): LongInt;
  begin
    Result := StrToIntDef(ReadString(section, key, IntToStr(def)), def);
  end;

  {
    Write a string value.

    @param section section name
    @param key key name
    @returns true, if value could be written to the object,
             false, if not (out of memory)
  }
  function TPrivateIniFile.putString(const section, key, value: String): Boolean;
  var
    sl : TObservedStringList;
    ts : TString;
    i : Integer;
  begin
    Result := false;
    // When no data is given, flush data; simulates the behaviour of
    // WriteProfileString
    if Length(section) + Length(key) + Length(value) = 0 then begin
      Flush;
      Result := True;
      exit;
    end;
    // Find section
    if Sections.Find(section, i) then begin
      // Get string list for the section
      sl := TObservedStringList(Sections.Objects[i]);
    end else begin
      // If section does not exist yet, create it
      sl := CreateStringList;
      if sl <> nil then begin
        try
          // Add section to list of sections
          Sections.AddObject(section, sl);
          oDirty := True;
        except
          on EOutOfMemory do begin
            sl.Free;
            sl := nil;
          end;
        end;
      end;
    end;
    if sl <> nil then begin
      // Delete old value, if one exist
      if sl.Find(key, i) then begin
        ts := TString(sl.Objects[i]);
        sl.Delete(i);
        ts.Free;
      end;
      // Add value to section
      try
        ts := TString.Create(value);
        try
          sl.AddObject(key, ts);
          oDirty := True;
          Result := True;
        except
          on EOutOfMemory do ts.Free;
        end;
      except
        on EInOutError do;
      end;
    end;
  end;

  {
    Write an integer value.

    @param section section name
    @param key key name
    @returns true, if value could be written to the object,
             false, if not (out of memory)
  }
  function TPrivateIniFile.putInt(const section, key: String;
      const value: Integer): Boolean;
  begin
    Result := putString(section, key, IntToStr(value));
  end;

  {
    Get a list of all keys of a section.

    @param section section name
    @param keys the list, where the keys to put in
  }
  procedure TPrivateIniFile.ReadSection(const section: String; keys: TStrings);
  var
    i : Integer;
    sl : TObservedStringList;
  begin
    if keys = nil then exit;
    if Sections.Find(section, i) then begin
      // Get string list of the section
      sl := TObservedStringList(Sections.Objects[i]);
      if sl = nil then exit;
      keys.BeginUpdate;
      for i := 0 to sl.Count-1 do begin
        try
          keys.Add(sl.Strings[i]);
        except
          on EOutOfMemory do exit;
        end;
      end;
      keys.EndUpdate;
    end;
  end;

  {
    Removes a complete section.

    @param section section name
  }
  procedure TPrivateIniFile.EraseSection(const section: String);
  var
    i : Integer;
    sl : TObservedStringList;
    ts : TString;
  begin
    if Sections.Find(section, i) then begin
      // Get String list of this section
      sl := TObservedStringList(Sections.Objects[i]);
      Sections.Delete(i);
      if sl <> nil then begin
        // Remove all values and free them
        sl.BeginUpdate;
        for i := 0 to sl.Count-1 do begin
          ts := TString(sl.Objects[i]);
          ts.Free;
          // We won't block the system
          if (i mod 100) = 99 then Application.ProcessMessages;
        end;
        sl.EndUpdate;
        sl.Clear;
      end;
      sl.Free;
      oDirty := True;
    end;
  end;

  {
    Removes a key and its associated value from a section
  }
  procedure TPrivateIniFile.DeleteKey(const section, key: String);
  var
    i, j : Integer;
    sl : TObservedStringList;
    ts : TString;
  begin
    if Sections.Find(section, i) then begin
      // Get string list of the section
      sl := TObservedStringList(Sections.Objects[i]);
      // Find key, remove it and free the value
      if sl.Find(key, j) then begin
        ts := TString(sl.Objects[j]);
        sl.delete(j);
        ts.Free;
        // Remove section, if it is now empty
        if sl.Count = 0 then begin
          Sections.Delete(i);
          sl.Free;
        end;
        oDirty := True;
      end;
    end;
  end;

  {
    Get a list of sections.

    @param sectlist the list where the sections names to put in
  }
  procedure TPrivateIniFile.ReadSections(sectlist: TStrings);
  var
    i: Integer;
  begin
    if sectlist = nil then exit;
    sectlist.BeginUpdate;
    for i := 0 to Sections.Count-1 do begin
      try
        sectlist.Add(Sections.Strings[i]);
      except
        on EOutOfMemory do exit;
      end;
    end;
    sectlist.EndUpdate;
  end;

  {
    Get a list of all values of a section.

    @param section the section name
    @param values the list, where the values to put in
  }
  procedure TPrivateIniFile.ReadSectionValues(const section: String;
      values: TStrings);
  var
    i : Integer;
    sl : TObservedStringList;
  begin
    if values = nil then exit;
    if Sections.Find(section, i) then begin
      // Get string list of section
      sl := TObservedStringList(Sections.Objects[i]);
      if sl = nil then exit;
      sl.BeginUpdate;
      for i := 0 to sl.Count-1 do begin
        try
          values.Add(TString(sl.Objects[i]).Value);
        except
          on EOutOfMemory do exit;
        end;
      end;
      sl.EndUpdate;
    end;
  end;

  {
    Write an integer value.

    @param section the section name
    @param key the key name
    @param value the value
  }
  procedure TPrivateIniFile.WriteInteger(const section, key: String;
      const value: LongInt);
  begin
    putInt(section, key, value);
  end;

  {
    Write a string value.

    @param section the section name
    @param key the key name
    @param value the value
  }
  procedure TPrivateIniFile.WriteString(const section, key, value: String);
  begin
    putString(section, key, value);
  end;

  {
    Flush changed values to file.
    If a filename is defined, this function writes all data to this file.
  }
  procedure TPrivateIniFile.Flush;
  begin
    if oDirty and (Length(oFileName) > 0) then Save(oFileName);
  end;

end.
