{
    Respond, the fax number form daemon
    Copyright (C) 2000 Horst F <horstf@gmx.de>

    This program is free software; you can redistribute it and/or modify
    it under the terms of the GNU General Public License as published by
    the Free Software Foundation; either version 2 of the License, or
    (at your option) any later version.

    This program is distributed in the hope that it will be useful,
    but WITHOUT ANY WARRANTY; without even the implied warranty of
    MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
    GNU General Public License for more details.

    You should have received a copy of the GNU General Public License
    along with this program; if not, write to the Free Software
    Foundation, Inc., 59 Temple Place, Suite 330, Boston, MA  02111-1307  USA
}
unit Observer;

interface

  uses
    classes;

  type
    {
      This class should prevent memory holes by registering every object on
      create and deregistering it on destroy.
      To register an object, Observe must be called. To unregister an object,
      Forget must be called.
    }
    TObjectObserver = class(Tobject)
    private
      // The list of registered objects
      oObjectList: TList;
      oObserve: Boolean;
      function GetCount: integer;
      function GetObject(const index: Integer): TObject;
    public
      constructor Create;
      destructor Destroy; override;
      procedure Observe(const anobject: TObject);
      procedure Forget(const anobject: TObject);
      function CountClassObjects(const aclass: TClass): Integer;
      procedure FreeAll;
      // Number of registered objects.
      property Count: Integer read GetCount;
      // Provides access to the list of registered objects.
      property Objects[const index: Integer]: TObject read GetObject;
    end;

  function ObjectObserver: TObjectObserver;

implementation

  var
    // The systemwide object observer
    cObjectObserver: TObjectObserver;
    SavedExitProc: Pointer;

  {
    Initializes the object.
  }
  constructor TObjectObserver.Create;
  begin
    oObjectList := nil;
    inherited Create;
    oObjectList := TList.Create;
{$IFDEF DEBUG}
    oObserve := True;
{$ELSE}
    oObserve := False;
{$ENDIF}
  end;

  {
    Cleanup of the object.
  }
  destructor TObjectObserver.Destroy;
  begin
    oObjectList.Free;
    inherited Destroy;
  end;

  {
    Returns the number of registered objects.

    @returns the number of registered objects
  }
  function TObjectObserver.GetCount: Integer;
  begin
    Result := oObjectList.Count;
  end;

  {
    Returns an registered object.

    @param index index of the object in the list
    @returns the requested object
  }
  function TObjectObserver.GetObject(const index: Integer): TObject;
  begin
    Result := oObjectList.Items[index];
  end;

  {
    Register an object.

    @param anobject the object to register
  }
  procedure TObjectObserver.Observe(const anobject: TObject);
  begin

    if oObserve then with oObjectList do begin
      // An object will only be registered once
      if IndexOf(anobject) = -1 then begin
        Add(anobject);
      end;
    end;
  end;

  {
    Unregister an object.

    @param anobject the object to unregister
  }
  procedure TObjectObserver.Forget(const anobject: TObject);
  var
    i: Integer;
  begin
    if oObserve then with oObjectList do begin
      i := IndexOf(anobject);
      if i <> -1 then begin
        Delete(i);
      end;
    end;
  end;

  {
    Counts all objects of the given class.

    @param aclass the class to count
    @returns the number of registered objects of the given class
  }
  function TObjectObserver.CountClassObjects(const aclass: TClass): Integer;
  var
    i: Integer;
  begin;
    Result := 0;
    for i := 0 to oObjectList.Count-1 do begin
      if TObject(oObjectList.Items[i]).ClassType = aclass then begin
        Inc(Result);
      end;
    end;
  end;

  procedure TObjectObserver.FreeAll;
  var
    i: Integer;
  begin
    for i := 0 to oObjectList.Count-1 do begin
      TObject(oObjectList.Items[i]).Free;
    end;
    oObjectList.Clear;
  end;

  {
    Returns the system wide object observer object.

    @returns the object observer
  }
  function ObjectObserver: TObjectObserver;
  begin
    Result := cObjectObserver;
  end;

  {
    Unit exit procedure.
  }
  procedure UnitExitProc;
  begin
    ExitProc := SavedExitProc;
    cObjectObserver.Free;
    cObjectObserver := nil;
  end;

begin
  cObjectObserver := TObjectObserver.Create;
  SavedExitProc := ExitProc;
  ExitProc := @UnitExitProc;
end.
